/*
LDAPEOAdaptor.framework - Providing EOF with an interface for LDAP
Copyright (C) 1998  Ryan Krebs and Stanford University

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Comments, suggestions, bug reports, etc can be directed to:
rkrebs@cisnet.stanford.edu
*/

#import "LDAPAdaptorPrivate.h"

/*  The LDAPChannel works with the LDAPController to
    perform searches, modifications, and deletions.
    The idea was to leave a class that could be used
    in later projects, but I'm not sure how well
    that's going to work.  */
@implementation LDAPChannel
// returns _isOpen
- (BOOL)isOpen
{
	return _isOpen;
}

// returns _fetchInProgress
- (BOOL)isFetchInProgress
{
	return _fetchInProgress;
}

// notifies its context, sets _isOpen, sets _controller
- (void)openChannel
{
	if (_isOpen) return;

	[(LDAPContext *)_context channelWillOpen];

	_isOpen = YES;
	_controller = [[(LDAPContext *)_context ldapController] retain];
}

// cancels fetch if necessary, notifies context, sets _isOpen and _controller
- (void)closeChannel
{
	if (!_isOpen) return;
	if (_fetchInProgress) [self cancelFetch];

	[_controller release];
	_controller = nil;

	_isOpen = NO;

	[(LDAPContext *)_context channelDidClose];
}

// sets _attributes and informs _controller of the change
- (void)setAttributesToFetch:(NSArray *)attributes
{
	[_attributes autorelease];
	_attributes = [attributes retain];

	[_controller setAttributesToFetch:attributes];
}

// returns _attributes
- (NSArray *)attributesToFetch
{
	return _attributes;
}

/* Is this really needed?  It would be nice to have the
   adaptor automatically assign the dn based on a couple
   other things, but that could be tricky.  */
- (NSString *)primaryKeyForNewRowWithEntity:(EOEntity *)entity
{
	return nil;
}

///////////////
// Operations
///////////////

// sets _attributes, gets search filter, starts search, sets _ldapMsgID
- (void)selectAttributes:(NSArray *)attributes fetchSpecification:(EOFetchSpecification *)fetchSpec lock:(BOOL)lock entity:(EOEntity *)entity
{
	NSString *filter;

	if (_delegateRespondsTo.shouldSelectAttributes) {
		if (![_delegate adaptorChannel:self shouldSelectAttributes:attributes fetchSpecification:fetchSpec lock:lock entity:entity]) {
			return;
		}
	}

	[self setAttributesToFetch:attributes];

	filter = [NSString stringWithString:[fetchSpec ldapFilterWithValueSeparator:[_controller valueSeparator]]];

	_ldapMsgID = 0;
	_ldapMsgID = [_controller initiateSearchWithFilter:filter];

	if (!_ldapMsgID)
		[NSException raise:EOGeneralAdaptorException format:@"The LDAPController was unable to begin a search in initiateSearchWithFilter:"];

	[(LDAPContext *)_context channelWillBeginFetching];
	_fetchInProgress = YES;

	if (_delegateRespondsTo.didSelectAttributes)
		[_delegate adaptorChannel:self didSelectAttributes:attributes fetchSpecification:fetchSpec lock:lock entity:entity];

	return;
}

// fetches a single row from _controller
- (NSMutableDictionary *)fetchRowWithZone:(NSZone *)zone
{
	LDAPMessage *entry;
	int errorNum;
	NSMutableDictionary *row;
	row = [NSMutableDictionary dictionaryWithCapacity:1];

	if (_delegateRespondsTo.willFetchRow) {
		[_delegate adaptorChannelWillFetchRow:self];
	}

	if (!_fetchInProgress) return nil;

// This isn't the most graceful way to fetch a result, but.....
	errorNum = [_controller retrieveResultsForID:_ldapMsgID all:NO message:&entry];
	if (errorNum == LDAP_RES_SEARCH_RESULT) {
		_fetchInProgress = NO;
		[(LDAPContext *)_context channelDidEndFetching];
		return nil;
	}
	if (errorNum != LDAP_RES_SEARCH_ENTRY) {
		[NSException raise:EOGeneralAdaptorException format:@"Did not receive expected results from LDAPController in fetchRowWithZone: - LDAPController replied: %@", [_controller errorString]];
	}

	row = [_controller dictionaryWithEntry:entry];

	if (row && _delegateRespondsTo.didFetchRow)
		[_delegate adaptorChannel:self didFetchRow:row];

	return row;
}

// tells _controller to abandon the fetch operation
- (void)cancelFetch
{
	int errorNum;

	if (!_fetchInProgress) return;

	errorNum = [_controller abandonOperationWithID:_ldapMsgID];

	if (errorNum != LDAP_SUCCESS)
		[NSException raise:EOGeneralAdaptorException format:@"Error cancelling fetch - LDAPController says: %@", [_controller errorString]];

	_ldapMsgID = 0;
	_fetchInProgress = NO;

	[(LDAPContext *)_context channelDidEndFetching];
}

// tells _controller to insert a row
- (void)insertRow:(NSDictionary *)row forEntity:(EOEntity *)entity
{
	NSString *newDN = [NSString stringWithString:[row objectForKey:@"dn"]];

	if (!newDN)
		newDN = [NSString stringWithString:[self primaryKeyForNewRowWithEntity:entity]];
	if (!newDN)
		[NSException raise:EOGeneralAdaptorException format:@"insertRow:forEntity: - Tried to add a new entry with no DN!"];

	if (![_controller addEntry:newDN withDictionary:row])
		[NSException raise:EOGeneralAdaptorException format:@"Error adding entry %@ - LDAPController says: $@", newDN, [_controller errorString]];

	return;
}

// tells _controller to update values and returns the number of entries updated
- (unsigned)updateValues:(NSDictionary *)values inRowsDescribedByQualifier:(EOQualifier *)qualifier entity:(EOEntity *)entity
{
	int result = [_controller updateEntriesMatching:qualifier withDictionary:values];
	if (!result)
		[NSException raise:EOGeneralAdaptorException format:@"Could not update entries - LDAPController says: %@", [_controller errorString]];

	return result;
}

// tells _controller to delete rows and returns the number of entries deleted
- (unsigned)deleteRowsDescribedByQualifier:(EOQualifier *)qualifier entity:(EOEntity *)entity
{
	int result = [_controller deleteEntriesMatching:qualifier];
	if (!result)
		[NSException raise:EOGeneralAdaptorException format:@"Could not delete entries - LDAPController says: %@", [_controller errorString]];

	return result;
}

////////////////
// Descriptions
////////////////

/* supposed to return the attributes which were fetched.
   The LDAPController supplies @"" for attributes that
   did not exist in the entry but were specified by
   attributesToFetch, so as to avoid nasty crashing. */
- (NSArray *)describeResults
{
	return _attributes;
}

/* searches for objectclasses in the entry at schemabase.
   (schemabase is set in the login panel, defaults to cn=schema).
   This must be an entry of objectclass=subschema, and follow
   the syntax that Netscape's Directory Server uses. There's
   probably an X.500 or LDAP RFC that sets down the syntax.
   You don't *need* the schema information from the server, but
   it makes building a model in EOModeler soooo much easier.

   Any suggestions for optimizations here are very welcome. */
- (NSArray *)describeTableNames
{
	NSString *searchBase = [NSString stringWithString:[_controller searchBase]];
	NSArray *attributes = [_controller attributesToFetch];
	int msgID, i, searchScope = [_controller searchScope];
	LDAPMessage *entry;
	char **values;
	NSArray *objectclasses;
	NSMutableArray *names = [NSMutableArray array];

	[_controller setSearchBase:[[(LDAPAdaptor *)[(LDAPContext *)_context adaptor] connectionDictionary] objectForKey:schemaBaseKey]];
	[_controller setAttributesToFetch:nil];
	[_controller setSearchScope:LDAP_SCOPE_BASE];

	msgID = [_controller initiateSearchWithFilter:@"(objectclass=subschema)"];

	if (!([_controller retrieveResultsForID:msgID all:NO message:&entry] == LDAP_RES_SEARCH_ENTRY))
		return nil;

	values = ldap_get_values ( [_controller connection], entry, (char *)"objectclasses" );
	objectclasses = [NSArray arrayWithCStringArray:values];
	ldap_value_free( values );

// All we're interested in right now is the table names, so that's what we'll scan for
	for (i=0; i<[objectclasses count]; i++) {
		NSString *tempName;
		NSScanner *scanner = [NSScanner scannerWithString:[objectclasses objectAtIndex:i]];
		[scanner setCaseSensitive:YES];
		if ([scanner scanUpToString:@"NAME '" intoString:NULL] &&
		[scanner scanString:@"NAME '" intoString:NULL] &&
		[scanner scanUpToString:@"'" intoString:&tempName])
			[names addObject:tempName];
	}

	[_controller setSearchBase:searchBase];
	[_controller setAttributesToFetch:attributes];
	[_controller setSearchScope:searchScope];

	if ([_controller retrieveResultsForID:msgID all:NO message:&entry] != LDAP_RES_SEARCH_RESULT)
		[NSException raise:NSInternalInconsistencyException format:@"Schema search returned more than one entry - describeTableNames"];

	if ( [names count] != 0 )
		return names;
	else
		return nil;
}

// searches for schema information for the tables and builds a model
// This code is really really messy
- (EOModel *)describeModelWithTableNames:(NSArray *)tableNames
{
	EOModel *model;
	EOAdaptor *adaptor = (LDAPAdaptor *)[(LDAPContext *)[self adaptorContext] adaptor];
	int msgID;
	LDAPMessage *entry;
	unsigned i,c;
	char **values;
	NSArray *attributetypes, *objectclasses;
	NSMutableDictionary *atDict, *ocDict;
// To restore old _controller values
	NSString *searchBase = [NSString stringWithString:[_controller searchBase]];
	NSArray *attributes = [NSArray arrayWithArray:[_controller attributesToFetch]];
	int searchScope = [_controller searchScope];

// top needs to be in there for relationships and stuff
	if (![tableNames containsObject:@"top"])
		tableNames = [tableNames arrayByAddingObject:@"top"];

// Next we'll retrieve the schema entry
	[_controller setSearchBase:[[(LDAPAdaptor *)[(LDAPContext *)_context adaptor] connectionDictionary] objectForKey:schemaBaseKey]];
	[_controller setAttributesToFetch:nil];
	[_controller setSearchScope:LDAP_SCOPE_BASE];

	msgID = [_controller initiateSearchWithFilter:@"(objectclass=subschema)"];
	if (!([_controller retrieveResultsForID:msgID all:NO message:&entry] == LDAP_RES_SEARCH_ENTRY))
		return nil;

	values = ldap_get_values ( [_controller connection], entry, (char *)"attributetypes" );
	attributetypes = [NSArray arrayWithCStringArray:values];
	ldap_value_free( values );
	values = ldap_get_values ( [_controller connection], entry, (char *)"objectclasses" );
	objectclasses = [NSArray arrayWithCStringArray:values];
	ldap_value_free( values );

// Stick all available attribute types and their syntaxes into atDict
	atDict = [NSMutableDictionary dictionary];
	for (i=0; i<[attributetypes count]; i++) {
		NSString *tempName, *tempSyntax;
		NSScanner *scanner = [NSScanner scannerWithString:[attributetypes objectAtIndex:i]];
		[scanner setCaseSensitive:YES];
// attributetypes have the form ( oid NAME 'name' DESC 'desc' SYNTAX 'syntax' [SINGLE-VALUE] )
		if ([scanner scanUpToString:@"NAME '" intoString:NULL] &&
			[scanner scanString:@"NAME '" intoString:NULL] &&
			[scanner scanUpToString:@"'" intoString:&tempName] &&
			[scanner scanUpToString:@"SYNTAX '" intoString:NULL] &&
			[scanner scanString:@"SYNTAX '" intoString:NULL] &&
			[scanner scanUpToString:@"'" intoString:&tempSyntax] )
		{
// The next part takes a guess at the type.
// This doesn't have to be perfect because the user can change the model in EOModeler
			if ([tempSyntax hasSuffix:@".5"] || [tempSyntax hasSuffix:@".4"] ||
				[tempSyntax hasSuffix:@".8"] || [tempSyntax hasSuffix:@".9"] ||
				[tempSyntax hasSuffix:@".10"] || [tempSyntax hasSuffix:@".23"] ||
				[tempSyntax hasSuffix:@".28"])
				[atDict setObject:@"Data" forKey:tempName];
			else
				[atDict setObject:@"String" forKey:tempName];
		}
	}

// Here's the fun part: the objectclasses
	ocDict = [NSMutableDictionary dictionary];
	for (i=0; i<[objectclasses count]; i++) {
		NSMutableDictionary *tempDict = [NSMutableDictionary dictionary];
		NSString *tempName, *tempValue;
		NSMutableArray *mustValues = [NSMutableArray array], *mayValues = [NSMutableArray array];
		NSScanner *scanner = [NSScanner scannerWithString:[objectclasses objectAtIndex:i]];
		[scanner setCaseSensitive:YES];
// objectclasses have the form ( oid NAME 'name' DESC 'desc' [SUP 'parent'] [MUST (value $ value)] [MAY (value $ value)] )
// we're going to ignore SUP and make top the parent for the purposes of the model.
// this isn't necessarily true, but it will make it easier to put in the dn, rdn, and parent stuff
		if (!([scanner scanUpToString:@"NAME '" intoString:NULL] &&
			[scanner scanString:@"NAME '" intoString:NULL] &&
			[scanner scanUpToString:@"'" intoString:&tempName]))
			continue;
		[tempDict setObject:tempName forKey:@"NAME"];

// getting MUSTs and MAYs in here, used to determine "allows NULL" in EOModeler.
		[scanner setScanLocation:0];
		[scanner scanUpToString:@"MUST (" intoString:NULL];
		if (![scanner isAtEnd]) {
			NSScanner *mustScanner;
			NSString *mustString;
			[scanner scanString:@"MUST (" intoString:NULL];
			[scanner scanUpToString:@" )" intoString:&tempValue];
			mustScanner = [NSScanner scannerWithString:tempValue];
			while (![mustScanner isAtEnd]) {
				[mustScanner scanUpToString:@" $" intoString:&mustString];
				[mustValues addObject:mustString];
				[mustScanner scanString:@"$" intoString:NULL];
			}
		}
		[tempDict setObject:mustValues forKey:@"MUST"];

		[scanner setScanLocation:0];
		[scanner scanUpToString:@"MAY (" intoString:NULL];
		if (![scanner isAtEnd]) {
			NSScanner *mayScanner;
			NSString *mayString;
			[scanner scanString:@"MAY (" intoString:NULL];
			[scanner scanUpToString:@" )" intoString:&tempValue];
			mayScanner = [NSScanner scannerWithString:tempValue];
			while (![mayScanner isAtEnd]) {
				[mayScanner scanUpToString:@" $" intoString:&mayString];
				[mayValues addObject:mayString];
				[mayScanner scanString:@"$" intoString:NULL];
			}
		}
		[tempDict setObject:mayValues forKey:@"MAY"];

		[ocDict setObject:tempDict forKey:tempName];
	}

// Now we actually start building the model.  Told you it was messy.

	model = [[EOModel new] autorelease];
	[model setAdaptorName:[adaptor name]];
	[model setConnectionDictionary:[adaptor connectionDictionary]];

// Need to step through names and build EOEntities based on ocDict and atDict here
	for(i=0; i<[tableNames count]; i++) {
		NSString *table = [tableNames objectAtIndex:i];
		EOEntity *entity;
		NSDictionary *entityDict = [ocDict objectForKey:table];
		NSArray *mustValues = [entityDict objectForKey:@"MUST"], *mayValues = [entityDict objectForKey:@"MAY"];
                EOAttribute *dnAttribute = [[EOAttribute alloc] init],
                        *rdnAttribute = [[EOAttribute alloc] init],
                        *parentAttribute = [[EOAttribute alloc] init];

// One would think it'd make more sense to do this outside the for loop, but
// If I try adding the same attribute to two entities, it gets unhappy.
// Even if the attributes are identical, they must be allocated separately.
                [dnAttribute setName:@"dn"];
                [dnAttribute setColumnName:@"dn"];
                [dnAttribute setExternalType:@"String"];
                [dnAttribute setValueClassName:@"NSString"];
                [dnAttribute setAllowsNull:NO];

                [parentAttribute setName:@"parent"];
                [parentAttribute setColumnName:@"parent"];
                [parentAttribute setExternalType:@"String"];
                [parentAttribute setValueClassName:@"NSString"];
                [parentAttribute setAllowsNull:YES];

                [rdnAttribute setName:@"rdn"];
                [rdnAttribute setColumnName:@"rdn"];
                [rdnAttribute setExternalType:@"String"];
                [rdnAttribute setValueClassName:@"NSString"];
                [rdnAttribute setAllowsNull:NO];

		entity = [[EOEntity allocWithZone:[model zone]] init];
		[entity setName:table];
		[entity setExternalName:table];
		[model addEntity:entity];
		[entity release];

// adding dn attribute for each entity's primary key
		[entity addAttribute:dnAttribute];
		if (![entity setPrimaryKeyAttributes:[NSArray arrayWithObject:dnAttribute]])
			[NSException raise:EOGeneralAdaptorException format:@"Can't set dn as primary key for table %@", table];

		[entity addAttribute:parentAttribute];
		[entity addAttribute:rdnAttribute];

		for (c=0; c<[mustValues count]; c++) {
			EOAttribute *attribute;
			NSString *tempName = [mustValues objectAtIndex:c];
			NSString *tempType = [atDict objectForKey:tempName];
			if (!tempType)
				tempType = [NSString stringWithString:@"String"];
			attribute = [[EOAttribute alloc] init];
			if (![tempName hasSuffix:@";binary"])
				[attribute setName:tempName];
			else {
				NSMutableString *newName = [NSMutableString stringWithString:tempName];
				[newName replaceCharactersInRange:[newName rangeOfString:@";"] withString:@"_"];
				[attribute setName:newName];
			}
			[attribute setColumnName:tempName];
			[attribute setExternalType:tempType];
			[attribute setValueClassName:[LDAPAdaptor internalTypeForExternalType:tempType model:model]];
			[attribute setAllowsNull:NO];
                        if (![entity attributeNamed:[attribute name]])
				[entity addAttribute:attribute];
		}

		for (c=0; c<[mayValues count]; c++) {
			EOAttribute *attribute;
			NSString *tempName = [mayValues objectAtIndex:c];
			NSString *tempType = [atDict objectForKey:tempName];
// Going to temporarily ignore attributes with the ;binary suffix
			if (![tempName hasSuffix:@";binary"]) {
				if (!tempType)
					tempType = [NSString stringWithString:@"String"];
				attribute = [[EOAttribute alloc] init];
// EOF doesn't like semicolons, does it?
     //                   if (![tempName hasSuffix:@";binary"])
                                [attribute setName:tempName];
     /*                   else {
                                NSMutableString *newName = [NSMutableString stringWithString:tempName];
                                [newName replaceCharactersInRange:[newName rangeOfString:@";"] withString:@"_"];
                                [attribute setName:newName];
                        }  */
				[attribute setColumnName:tempName];
				[attribute setExternalType:tempType];
				[attribute setValueClassName:[LDAPAdaptor internalTypeForExternalType:tempType model:model]];
				[attribute setAllowsNull:YES];
				if (![entity attributeNamed:[attribute name]])
					[entity addAttribute:attribute];
			}
		}

	}

// setting up relationships
	for (i=0; i<[tableNames count]; i++) {
		EOEntity *top = [model entityNamed:@"top"];
		EOEntity *child = [model entityNamed:[tableNames objectAtIndex:i]];

		EOAttribute *topDNAttribute = [top attributeNamed:@"dn"];
		EOAttribute *topParentAttribute = [top attributeNamed:@"parent"];
		EOAttribute *dnAttribute = [child attributeNamed:@"dn"];
		EOAttribute *parentAttribute = [child attributeNamed:@"parent"];

		EOJoin *toParentJoin = [[[EOJoin alloc] initWithSourceAttribute:parentAttribute destinationAttribute:topDNAttribute] autorelease];
		EOJoin *toChildrenJoin = [[[EOJoin alloc] initWithSourceAttribute:dnAttribute destinationAttribute:topParentAttribute] autorelease];

		EORelationship *toParent = [[[EORelationship alloc] init] autorelease];
		EORelationship *toChildren = [[[EORelationship alloc] init] autorelease];

		[toParent setName:@"toParent"];
		[child addRelationship:toParent];
		[toParent addJoin:toParentJoin];
		[toParent setToMany:NO];
		[toParent setJoinSemantic:EOInnerJoin];

		[toChildren setName:@"toChildren"];
		[child addRelationship:toChildren];
		[toChildren addJoin:toChildrenJoin];
		[toChildren setToMany:YES];
		[toChildren setJoinSemantic:EOInnerJoin];
	}

// setting up inheritance
	for (i=0; i<[tableNames count]; i++)
		if (![[tableNames objectAtIndex:i] isEqualToString:@"top"]) {
			EOEntity *top = [model entityNamed:@"top"];
			EOEntity *child = [model entityNamed:[tableNames objectAtIndex:i]];
			[top addSubEntity:child];
		}


// restore old values
	[_controller setSearchBase:searchBase];
	[_controller setAttributesToFetch:attributes];
	[_controller setSearchScope:searchScope];

	return model;
}
@end
