/*
LDAPEOAdaptor.framework - Providing EOF with an interface for LDAP
Copyright (C) 1998  Ryan Krebs and Stanford University

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Comments, suggestions, bug reports, etc can be directed to:
rkrebs@cisnet.stanford.edu
*/

#import "LDAPAdaptorPrivate.h"

@implementation LDAPContext
// Self explanatory
- (EOAdaptorChannel *)createAdaptorChannel
{
	return [[[LDAPChannel alloc] initWithAdaptorContext:self] autorelease];
}

/*  It seems like something more should be done for the begin-, commit-,
    and rollback- methods here, but since LDAP isn't quite a database,
    I'm not really sure what should happen.  */

// Makes sure a transaction is not already in progress, etc
- (void)beginTransaction
{
	if (!_controller)
		[NSException raise:NSInternalInconsistencyException format:@"Attempted to begin a transaction without connecting"];

	if (_transactionNestingLevel)
		[NSException raise:NSInternalInconsistencyException format:@"Attempted to nest a transaction"];

        if (_delegateRespondsTo.shouldBegin) {
                if(![_delegate adaptorContextShouldBegin:self])
                        return;
        }

        [self transactionDidBegin];

        if (_delegateRespondsTo.didBegin)
                [_delegate adaptorContextDidBegin:self];
}

- (void)commitTransaction
{
	if (!_transactionNestingLevel)
		[NSException raise:NSInternalInconsistencyException format:@"Tried to commit a transaction without beginning one"];

	if (_fetchesInProgress)
		[NSException raise:NSInternalInconsistencyException format:@"Attempted to commit a transaction during a fetch"];

        if (_delegateRespondsTo.shouldCommit) {
                if(![_delegate adaptorContextShouldCommit:self])
                        return;
        }

        [self transactionDidCommit];

        if (_delegateRespondsTo.didCommit)
                [_delegate adaptorContextDidCommit:self];
}

- (void)rollbackTransaction
{
	if (!_transactionNestingLevel)
		[NSException raise:NSInternalInconsistencyException format:@"Tried to rollback a transaction without beginning one"];

	if (_fetchesInProgress)
		[NSException raise:NSInternalInconsistencyException format:@"Attempted to rollback a transaction during a fetch"];

        if (_delegateRespondsTo.shouldRollback) {
                if(![_delegate adaptorContextShouldRollback:self])
                        return;
        }

        [self transactionDidRollback];

        if (_delegateRespondsTo.didRollback)
                [_delegate adaptorContextDidRollback:self];
}

// returns NO.  Couldn't you tell?
- (BOOL)canNestTransactions
{
	return NO;
}

// returns _transactionNestingLevel, hopefully 0 or 1
- (unsigned)transactionNestingLevel
{
	return _transactionNestingLevel;
}

// returns the controller so the LDAPChannel can grab it
- (LDAPController *)ldapController
{
	return _controller;
}

// sets the controller... the method above didn't seem complete without it
- (void)setLdapController:(LDAPController *)controller
{
	if (_controller)
		[_controller release];
	_controller = [controller retain];
}

// disconnects if necessary, then calls super
- (void)dealloc
{
	if (_controller) {
		[_controller release];
		_controller = nil;
	}

	[super dealloc];
}

//////////////////////
// Channel Notifiers
//////////////////////

// if not connected, it will connect
- (void)channelWillOpen
{
	if (!_openChannelCount) {
		_controller = [[LDAPController controllerWithConnectionDictionary:[(LDAPAdaptor *)_adaptor connectionDictionary]] retain];
		[_controller setShouldRaise:NO];
		[_controller connect];
	}
	_openChannelCount++;
}

// if this is the last channel, it will disconnect from the server
- (void)channelDidClose
{
	if (_openChannelCount) _openChannelCount--;
	if (!_openChannelCount) {
		[_controller release];
		_controller = nil;
	}
}

// incrementes fetch counter.  I love those little ++ things
- (void)channelWillBeginFetching
{
	_fetchesInProgress++;
}

// decrements fetch counter.  The --'s aren't quite as cool as the ++'s
- (void)channelDidEndFetching
{
	_fetchesInProgress--;
}
@end
