/*
LDAPEOAdaptor.framework - Providing EOF with an interface for LDAP
Copyright (C) 1998  Ryan Krebs and Stanford University

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Comments, suggestions, bug reports, etc can be directed to:
rkrebs@cisnet.stanford.edu
*/

#import "LDAPAdaptorPrivate.h"

static void freearray(char ** array)
{
        char *element = array[0];
        int i;

        for (i=0; element != NULL; i++, element = array[i])
                free(element);

        free(array);
}

static char *strdup(const char *s1)
{
        char *s;
        int len = strlen(s1);

        s = (char *)calloc(len+1, sizeof(char));
        if (!s) return NULL;
        strncpy(s, s1, len+1);
        s[len] = '\0';
        return s;
}

@implementation LDAPController
// I hope this is right.  What's involved in allocating an object?
+ (LDAPController *)controllerWithConnectionDictionary:(NSDictionary *)dictionary
{
	LDAPController *controller = [LDAPController alloc];
	[controller initWithConnectionDictionary:dictionary];
	[controller autorelease];

	return controller;
}

// Sets up all the stuff likely to be needed in an LDAP session
// using an NSDictionary with the required stuff
- (void)initWithConnectionDictionary:(NSDictionary *)dictionary
{
	NSString *temp;

	_searchScope = LDAP_SCOPE_SUBTREE;
	_errorNum = 0;

	_serverName = [NSString stringWithString:[dictionary objectForKey:serverNameKey]];
	if (!_serverName) [NSException raise:NSInvalidArgumentException format:@"The dictionary passed to LDAPController contains no server name.  (Make sure serverNameKey is defined.)"];

	_userDN = [NSString stringWithString:[dictionary objectForKey:userDNKey]];
	if (!_userDN) _userDN =[NSString stringWithString:@""];

	_searchBase = [NSString stringWithString:[dictionary objectForKey:defaultBaseKey]];
	if (!_searchBase) [NSException raise:NSInvalidArgumentException format:@"The dictionary passed to LDAPController contains no search base.  (Make sure defaultBaseKey is defined.)"];

	_password = [NSString stringWithString:[dictionary objectForKey:passwordKey]];
	if (!_password) _password = [NSString stringWithString:@""];

	_serverPort = [[dictionary objectForKey:serverPortKey] intValue];
	if (!_serverPort) _serverPort = 389;

	_timeout = [[dictionary objectForKey:timeoutKey] intValue];
	if (!_timeout) _timeout = 0;

	_valueSeparator = [dictionary objectForKey:valueSeparatorKey];
	if (!_valueSeparator) _valueSeparator = [NSString stringWithString:@" | "];

	temp = [NSString stringWithString:[dictionary objectForKey:authenticationMethodKey]];

	_authMethod = LDAP_AUTH_SIMPLE;

#ifdef LDAP_AUTH_KRBV4
	if ([temp isEqualToString:@"LDAP_AUTH_KRBV4"])
		_authMethod = LDAP_AUTH_KRBV4;
#endif

#ifdef LDAP_AUTH_SRP
// yeah, this would be nice.....
	if ([temp isEqualToString:@"LDAP_AUTH_SRP"])
		_authMethod = LDAP_AUTH_SRP;
#endif

// While there's a slim chance that somebody might, by default, want to
// fetch attribute names only, I think it's safe to do this here.
	_attributesOnly = NO;

// We're going to tell the controller to raise exceptions on errors,
// but the LDAPContext can tell it not to.
	_shouldRaise = YES;
}

// Set and get stuff here

// This keeps two copies, one is an array of attribute names
// which is used when performing searches.  The other is an
// array of EOAttributes, used to get the external type
// before fetching a value for the attribute.
- (void)setAttributesToFetch:(NSArray *)attributes
{
	int i, count;
	NSMutableArray *newAttributes = [NSMutableArray array];
	NSMutableArray *newEOAttributes = [NSMutableArray array];

        [_attributes autorelease];
	[_eoAttributes autorelease];

	if (!attributes) {
		_attributes = nil;
		_eoAttributes = nil;
	} else {
		count = [attributes count];
		for (i=0; i<count; i++)
			if (!([[[attributes objectAtIndex:i] columnName] isEqualToString:@"dn"] || [[[attributes objectAtIndex:i] columnName] isEqualToString:@"parent"] || [[[attributes objectAtIndex:i] columnName] isEqualToString:@"rdn"])) {
				[newAttributes addObject:[[attributes objectAtIndex:i] columnName]];
				[newEOAttributes addObject:[attributes objectAtIndex:i]];
			}

		_attributes = [newAttributes retain];
		_eoAttributes = [newEOAttributes retain];
	}
}

- (NSArray *)attributesToFetch
{
	return _attributes;
}

- (void)setAttributesOnly:(BOOL)yn
{
	_attributesOnly = yn;
}

- (BOOL)attributesOnly
{
	return _attributesOnly;
}

- (void)setSearchBase:(NSString *)base
{
	[_searchBase autorelease];
	_searchBase = [base retain];
}

- (NSString *)searchBase
{
	return _searchBase;
}

- (void)setSearchScope:(int)scope
{
	_searchScope = scope;
}

- (int)searchScope
{
	return _searchScope;
}

- (void)setValueSeparator:(NSString *)valSep
{
	[_valueSeparator autorelease];
	_valueSeparator = [valSep retain];
}

- (NSString *)valueSeparator
{
	return _valueSeparator;
}

- (void)setTimeout:(int)timeout
{
	_timeout = timeout;
}

- (int)timeout
{
	return _timeout;
}

- (void)setUserDN:(NSString *)userDN password:(NSString *)password {
	_userDN = userDN;
	_password = password;
}

- (LDAP *)connection
{
	return _connection;
}

// A handy little thing to ensure a connection before an operation
- (unsigned)autoconnect
{
	if (!_connection)
		if (![self connect]) {
			if (_shouldRaise)
				[NSException raise:EOGeneralAdaptorException format:@"Cannot connect to LDAP server to begin search - searchWithFilter:"];
			return 0;
		};

	return 1;
}

// Connect to the LDAP server and authenticate
- (unsigned)connect
{
	int errorNum;

	if (_connection) return 1;

	if ( (_connection = ldap_open((char *)[_serverName cString], _serverPort)) == NULL ) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Unable to connect to %@:%@ - ldap_open()", _serverName, _serverPort];
		return 0;
	}

	switch (_authMethod) {
#ifdef LDAP_AUTH_KRBV4
		case LDAP_AUTH_KRBV4:
			errorNum = ldap_kerberos_bind_s( _connection, (char *)[_userDN cString] );
			break;
#endif
#ifdef LDAP_AUTH_SRP
// We can dream, can't we?
		case LDAP_AUTH_SRP:
			errornum = ldap_srp_bind_s( _connection, (char *)[_userDN cString], (char *)[_password cString] );
			break;
#endif 
		case LDAP_AUTH_SIMPLE:
		default:
			errorNum = ldap_bind_s( _connection, (char *)[_userDN cString], (char *)[_password cString], _authMethod);
			break;
	}

	if (errorNum != LDAP_SUCCESS ) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Unable to authenticate with login: %@ - LDAP says: %@", _userDN, [self errorStringWithErrorNum:errorNum]];
		return 0;
	}
	return 1;
}

// Disconnect and get rid of the LDAP*
- (unsigned)disconnect
{
	int errorNum;

	if (!_connection) return 1;

	errorNum = ldap_unbind(_connection);
	if (errorNum != LDAP_SUCCESS) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Error disconnecting from LDAP server - LDAP says: %@", [self errorStringWithErrorNum:errorNum]];
		return 0;
	}
	_connection = NULL;
	return 1;
}

// Disconnect nicely if necessary
- (void)dealloc
{
	if (_connection)
		[self disconnect];
	[super dealloc];
}

// Tell the controller whether it should raise an exception on an error
- (void)setShouldRaise:(BOOL)raise
{
	_shouldRaise = raise;
}

- (BOOL)shouldRaise
{
	return _shouldRaise;
}

// Returns the string matching the LDAP error stored in _connection
- (NSString *)errorString
{
	int errno;
#ifdef LDAP_OPT_ON
// Netscape SDK
	errno = ldap_get_lderrno( _connection, NULL, NULL );
#else
	errno = _connection->ld_errno;
#endif
        return [self errorStringWithErrorNum:errno];
}

// Returns the string matching the LDAP error errorNum
- (NSString *)errorStringWithErrorNum:(int)errorNum
{
	NSMutableString *errorString = [NSMutableString stringWithFormat:@"Error number %d: ", errorNum];
	if (errorNum == LDAP_SUCCESS)
		[errorString appendString:@"Nothing wrong here... I bet the programmer screwed up someplace. :P "];
	else
		[errorString appendString:[NSString stringWithCString:ldap_err2string( errorNum )]];

	return errorString;
}

/////////////////////
// AsynchOperations
/////////////////////

// Start an asynchronous search with the filter and return the LDAP message ID
- (unsigned)initiateSearchWithFilter:(NSString *)filter
{
	int msgID;
	char **attributes;

	if (![self autoconnect]) return 0;

	if (!_attributes)
		attributes = NULL;
	else
		attributes = (char **)[_attributes cStringArray];

	msgID = ldap_search( _connection, (char *)[_searchBase cString], _searchScope, (char *)[filter cString], attributes, _attributesOnly);

	if (msgID < 0) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Error in initiateSearchWithFilter: - LDAP says: %@", [self errorString]];
		return 0;
	}

	if (attributes)
		freearray(attributes);

	return msgID;
}

// retrieves the results from a previous operation, storing search results in the LDAPMessage **
- (int)retrieveResultsForID:(int)msgID all:(BOOL)all message:(LDAPMessage **)message
{
	int result;

	if (_timeout) {
		struct timeval tv;
		tv.tv_sec = (unsigned long) _timeout;
		tv.tv_usec = 0;
		result = ldap_result( _connection, msgID, all, &tv, message);
	} else {
		result = ldap_result( _connection, msgID, all, NULL, message);
	}

	if (result < 1)
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Error retrieving results - LDAP says: %@", [self errorString]];

	return result;
}

// Retrieves results for a previous operation,
// most likely not a search since this one just discards the LDAPMessage *
- (int)retrieveResultsForID:(int)msgID
{
	LDAPMessage *result;
	return [self retrieveResultsForID:msgID all:YES message:&result];
}

- (unsigned)abandonOperationWithID:(int)msgID
{
	int errorNum;

	if (!_connection)
		return LDAP_SUCCESS;

	errorNum = ldap_abandon( _connection, msgID );
	if (errorNum != LDAP_SUCCESS)
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Error abandoning operation - LDAP says: %@", [self errorStringWithErrorNum:errorNum]];

	return errorNum;
}

///////////////////////
// SynchOperations
///////////////////////

// performs a synchronous search and returns an array of NSDictionary objects built from the results
- (NSArray *)searchWithFilter:(NSString *)filter
{
        char **attributes;
        int errorNum;
        LDAPMessage *result;

        if (![self autoconnect]) return 0;

        if (!_attributes)
                attributes = NULL;
        else
                attributes = (char **)[_attributes cStringArray];

        errorNum = ldap_search_s( _connection, (char *)[_searchBase cString], _searchScope, (char *)[filter cString], attributes, _attributesOnly, &result);

        if (errorNum != LDAP_SUCCESS) {
                if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Error searching for %@ - LDAP says: %@", filter, [self errorStringWithErrorNum:errorNum]];
                return [NSArray array];
        }

        return [self dictionaryArrayWithEntries:result];
}

// Synchronously deletes an entry, raising an exception if there is an error
- (unsigned)deleteEntry:(NSString *)dn
{
        int errorNum;

        if (![self autoconnect]) return 0;

        errorNum = ldap_delete_s( _connection, (char *)[dn cString] );

        if (errorNum != LDAP_SUCCESS) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Could not delete entry %@ - LDAP says: %@", dn, [self errorStringWithErrorNum:errorNum]];
                return 0;
        }
        return 1;
}

// Synchronously deletes entries which match an EOQualifier
- (unsigned)deleteEntriesMatching:(EOQualifier *)qualifier
{
        int i, count;
        NSArray *names = [self findEntriesMatching:qualifier];

        count = [names count];

        if (count)
                for (i = 0; i < count; i++) {
                        if (![self deleteEntry:[names objectAtIndex:i]])
                                return i;
                }

        return count;
}

// Converts the NSDictionary to an LDAPMod** and adds the entry
- (unsigned)addEntry:(NSString *)dn withDictionary:(NSDictionary *)dictionary
{
        int result;
        LDAPMod **mods = [self modificationsWithDictionary:dictionary];

        result = [self addEntry:dn withMods:mods];

        [LDAPController freeMods:mods];
        return result;
}

// Adds a new entry, raising an exception if there's an error
- (unsigned)addEntry:(NSString *)dn withMods:(LDAPMod **)mods
{
        int errorNum;

        errorNum = ldap_add_s( _connection, (char *)[dn cString], mods);

        if (errorNum != LDAP_SUCCESS) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Could not add entry %@ - LDAP says: %@", dn, [self errorStringWithErrorNum:errorNum]];
        	return 0;
	}
	return 1;
}

// Converts an NSDictionary to LDAPMod ** and updates the entry
- (unsigned)updateEntry:(NSString *)dn withDictionary:(NSDictionary *)dictionary
{
        int result;
	NSString *newRDN;
	LDAPMod **mods;

        mods = [self modificationsWithDictionary:dictionary];
        result = [self updateEntry:dn withMods:mods];

        [LDAPController freeMods:mods];

	if (result) {
		newRDN = [dictionary objectForKey:@"rdn"];

		if (newRDN) {
			if (![self renameEntry:dn withNewRDN:newRDN]) {
				if (_shouldRaise)
					[NSException raise:EOGeneralAdaptorException format:@"Could not rename entry %@ - LDAP says: %@", dn, [self errorString]];
				return 0;
			}
		}
	}
        return result;
}

// Updates an entry using an LDAPMod **, raising an exception on failure
- (unsigned)updateEntry:(NSString *)dn withMods:(LDAPMod **)mods
{
        int errorNum;

        errorNum = ldap_modify_s( _connection, (char *)[dn cString], mods);

        if (errorNum != LDAP_SUCCESS) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Could not update entry %@ - LDAP says: %@", dn, [self errorStringWithErrorNum:errorNum]];
		return 0;
	}
	return 1;
}

// Updates a set of entries matching an EOQualifier
// This finds entries matching and then calls updateEntry:withDictionary, because
// if I converted to mods and called updateEntriesMatching:withMods:, a modrdn could slip through
- (unsigned)updateEntriesMatching:(EOQualifier *)qualifier withDictionary:(NSDictionary *)dictionary
{
	int i, count;
	NSArray *names = [self findEntriesMatching:qualifier];

	count = [names count];

	if (count)
		for (i = 0; i < count; i++) {
			if (![self updateEntry:[names objectAtIndex:i] withDictionary:dictionary])
				return i;
		}

	return count;
}

// Updates entries matching an EOQualifier
- (unsigned)updateEntriesMatching:(EOQualifier *)qualifier withMods:(LDAPMod **)mods
{
        int i, count;
        NSArray *names = [self findEntriesMatching:qualifier];

        count = [names count];

        if (count)
                for (i = 0; i < count; i++) {
                        if (![self updateEntry:[names objectAtIndex:i] withMods:mods])
                                return i;
                }

        return count;
}

// Performs an asynchronous search to get the names of the entries matching an EOQualifier
- (NSArray *)findEntriesMatching:(EOQualifier *)qualifier
{
	int msgID;
	NSString *filter;
	LDAPMessage *result;
	NSMutableArray *names;
	char *name;

// Since we just want entry names, we'll temporarily set the controller to only fetch the c attribute name with the entry
	BOOL attrsOnly = _attributesOnly;
	NSArray *attributes = _attributes;
	_attributes = [NSArray arrayWithObject:@"c"];
	_attributesOnly = YES;

	filter = [NSString stringWithString:[qualifier ldapFilterWithValueSeparator:_valueSeparator]];
	msgID = [self initiateSearchWithFilter:filter];
	names = [NSMutableArray array];

	while ([self retrieveResultsForID:msgID all:NO message:&result] == LDAP_RES_SEARCH_ENTRY) {
		name = ldap_get_dn( _connection, result );
		[names addObject:[NSString stringWithCString:name]];
		ldap_memfree( name );
	}

	_attributesOnly = attrsOnly;
	_attributes = attributes;

	return names;
}

- (unsigned)renameEntry:(NSString *)dn withNewRDN:(NSString *)newRDN
{
	int errorNum;

	errorNum = ldap_modrdn2_s( _connection, (char *)[dn cString], (char *)[newRDN cString], 0 );

	if (errorNum != LDAP_SUCCESS) {
		if (_shouldRaise)
			[NSException raise:EOGeneralAdaptorException format:@"Could not rename entry %@ - LDAP says: %@", dn, [self errorStringWithErrorNum:errorNum]];
		return 0;
	}
	return 1;
}

////////////////
// Conversions
////////////////

// Gets the parent entry's dn
- (NSString *)parentOfEntryWithName:(NSString *)dn
{
	int i;
	char **rdns;
	NSMutableString *parent;

	rdns = ldap_explode_dn( (char *)[dn cString], 0 );
	if (!rdns)
		return [NSString stringWithString:@""];

	parent = [NSMutableString stringWithCString:rdns[1]];
	for (i = 2; rdns[i] != NULL; ++i)
		[parent appendFormat:@", %@", [NSString stringWithCString:rdns[i]]];

	ldap_value_free( rdns );

	return parent;
}

// Returns the first rdn from ldap_explode_dn()
- (NSString *)rdnOfEntryWithName:(NSString *)dn
{
	char **rdns;
	NSString *rdn;

	rdns = ldap_explode_dn( (char *)[dn cString], 0 );
	if (!rdns)
		return [NSString stringWithString:dn];

	rdn = [NSString stringWithCString:rdns[0]];

	ldap_value_free( rdns );
	return rdn;
}

// Builds an NSDictionary from an LDAPMessage*
// This method also adds the parent entry's dn under the key "parent"
// and adds an empty string for any attribute that was to be fetched but had no value
// The empty string allows the EOAdaptor to take a snapshot that includes attributes with no values
- (NSMutableDictionary *)dictionaryWithEntry:(LDAPMessage *)entry
{
	NSMutableDictionary *dict = [NSMutableDictionary dictionary];
	NSString *attribute;
	char *name;
	char **values;
	BerElement *ber;
	int i, count;

	name = ldap_get_dn( _connection, entry );
	[dict setObject:[NSString stringWithCString:name] forKey:@"dn"];

	[dict setObject:[self parentOfEntryWithName:[NSString stringWithCString:name]] forKey:@"parent"];
	[dict setObject:[self rdnOfEntryWithName:[NSString stringWithCString:name]] forKey:@"rdn"];

	ldap_memfree( name );

	if (!_attributes) {
		name = ldap_first_attribute( _connection, entry, &ber );
		do {
			NSArray *valArray;
			NSString *attribute = [NSString stringWithCString:name];

			values = ldap_get_values( _connection, entry, name );
			valArray = [NSArray arrayWithCStringArray:values];
			[dict setObject:[valArray componentsJoinedByString:_valueSeparator] forKey:attribute];
			ldap_value_free( values );
		} while ( name = ldap_next_attribute( _connection, entry, ber ) );
	} else {
		count = [_attributes count];
		for (i=0; i<count; i++) {
			EOAttribute *tempAttribute = [_eoAttributes objectAtIndex:i];
			attribute = [NSString stringWithString:[tempAttribute columnName]];

			if ([[tempAttribute externalType] isEqualToString:@"Data"]) {
				struct berval **values = ldap_get_values_len( _connection, entry, (char *)[attribute cString] );
				if ( ((struct berval **)values == NULL) || ((struct berval *)values[0] == NULL) )
					[dict setObject:[NSData data] forKey:attribute];
				else {
					int q;
					NSMutableArray *data = [NSMutableArray array];
					for (q=0; (struct berval *)values[q] != NULL; q++) {
						NSData *purityTestsSuck = [NSData dataWithBytes:values[q]->bv_val length:values[q]->bv_len];
						[data addObject:purityTestsSuck];
					}
					if ([data count] == 1)
						[dict setObject:[data objectAtIndex:0] forKey:attribute];
					else
						[dict setObject:data forKey:attribute];
					ldap_value_free_len( values );
				}

			} else  /* attribute is not binary */ {
				NSArray *valArray;
				values = ldap_get_values( _connection, entry, (char *)[attribute cString] );
				if (!values)
					[dict setObject:@"" forKey:attribute];
				else {
					valArray = [NSArray arrayWithCStringArray:values];
					[dict setObject:[valArray componentsJoinedByString:_valueSeparator] forKey:attribute];
					ldap_value_free( values );
				}
			}
		}
	}

	return dict;
}

// Simply runs through a chain of entries, building an NSDictionary for each and sticking them in an array
- (NSArray *)dictionaryArrayWithEntries:(LDAPMessage *)entries
{
	NSMutableArray *array;
	LDAPMessage *entry;

	array = [NSMutableArray arrayWithCapacity:ldap_count_entries( _connection, entries )];
	entry = ldap_first_entry( _connection, entries );
	while ( entry ) {
		[array addObject:[self dictionaryWithEntry:entry]];
		entry = ldap_next_entry( _connection, entry );
	}
	return array;
}

// Builds an LDAPMod *
// This has to check for deleted attributes, or else the empty attribute will later
// cause a seg fault when fetched.
- (LDAPMod *)modificationForAttribute:(NSString *)attribute values:(NSArray *)values
{
	LDAPMod *mod;
	int i, count;

	if (!values)
		count = 0;
	else
		count = [values count];

	mod = (LDAPMod *)malloc(sizeof(LDAPMod));
	assert(mod != NULL);

        mod->mod_type = (char *)[attribute cString];

	if (count) {
		mod->mod_op = LDAP_MOD_REPLACE;
		if ( [[values objectAtIndex:0] isKindOfClass:[NSData class]] ) {
			mod->mod_op |= LDAP_MOD_BVALUES;
			mod->mod_bvalues = (struct berval **)calloc(count+1, sizeof(struct berval));
			for (i=0; i<count; i++) {
				mod->mod_bvalues[i]->bv_val = (void *)[[values objectAtIndex:i] bytes];
				mod->mod_bvalues[i]->bv_len = [[values objectAtIndex:i] length];
			}
			mod->mod_bvalues[i] = NULL;
		} else {
			mod->mod_values = (char **)calloc(count+1, sizeof(char *));
                	for (i=0; i < count; i++) {
                		mod->mod_values[i] = strdup([[values objectAtIndex:i] cString]);
                	}
                	mod->mod_values[i] = NULL;
		}
	} else {
		mod->mod_op = LDAP_MOD_DELETE;
		mod->mod_values = NULL;
	}
	return mod;
}

// Builds a NULL terminated array of LDAPMod *s, excluding dn and parent if they get this far.
- (LDAPMod **)modificationsWithDictionary:(NSDictionary *)dictionary
{
	NSMutableDictionary *newDict = [NSMutableDictionary dictionaryWithDictionary:dictionary];
	LDAPMod **mods;
	NSArray *attributes;
	NSString *key, *values;
	int i, count;
	[newDict removeObjectForKey:@"dn"];
	[newDict removeObjectForKey:@"parent"];
	[newDict removeObjectForKey:@"rdn"];
	attributes = [newDict allKeys];
	count = [attributes count];

	mods = (LDAPMod **)calloc(count + 1, sizeof(LDAPMod *));
	assert(mods != NULL);

	for (i = 0; i < count; i++) {
		key = [NSString stringWithString:[attributes objectAtIndex:i]];
		values = [NSString stringWithString:[newDict objectForKey:key]];
		if ([values isEqualToString:@""])
			mods[i] = [self modificationForAttribute:key values:nil];
		else
			mods[i] = [self modificationForAttribute:key values:[values componentsSeparatedByString:_valueSeparator]];
	}
	mods[i] = NULL;
	return mods;
}

// There's a problem some place when the LDAPMods are allocated, such that simply
// calling ldap_mods_free() can cause a crash, while substituting the source code doesn't
// I'd love to know why this happens.
+ (void)freeMods:(LDAPMod **)mods
{
//	ldap_mods_free(mods, 1);
        int	i, j;

        if ( mods == NULL )
                return;

        for ( i = 0; mods[i] != NULL; i++ ) {
                if ( mods[i]->mod_op & LDAP_MOD_BVALUES ) {
                        ber_bvecfree( mods[i]->mod_bvalues );
                } else {
//                        ldap_value_free( mods[i]->mod_values );
                          if (mods[i]->mod_values != NULL) {
                                for (j=0; mods[i]->mod_values[j] != NULL; j++)
                                        free( mods[i]->mod_values[j] );
                                free( (char *) mods[i]->mod_values );
                          }
                }
                free( (char *) mods[i] );
        }
        free( (char *) mods ); 

}

// See comment above...
+ (void)freeMod:(LDAPMod *)mod
{
	LDAPMod *mods[2] = { mod, NULL };
//	ldap_mods_free(mods, 0);
	int	i, j;

	if ( mods == NULL )
        	return;

	for ( i = 0; mods[i] != NULL; i++ ) {
        	if ( mods[i]->mod_op & LDAP_MOD_BVALUES ) {
                	ber_bvecfree( mods[i]->mod_bvalues );
        	} else {
//	                ldap_value_free( mods[i]->mod_values );
                	if (mods[i]->mod_values != NULL) {
				for (j=0; mods[i]->mod_values[j] != NULL; j++)
					free( mods[i]->mod_values[j] );
				free( (char *) mods[i]->mod_values );
	                }

        	}
        	free( (char *) mods[i] );
	}
}
@end