/*
LDAPEOAdaptor.framework - Providing EOF with an interface for LDAP
Copyright (C) 1998  Ryan Krebs and Stanford University

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Comments, suggestions, bug reports, etc can be directed to:
rkrebs@cisnet.stanford.edu
*/

#import "LDAPAdaptorPrivate.h"
#import <string.h>

// Giving credit where due, I adapted freearray, strdup, and the cStringArray stuff
// from Luke Howard's LDAPAccess framework, since I hate pointers and malloc so much
// Thanks for the help Luke!

static void freearray(char ** array)
{
        char *element = array[0];
        int i;

        for (i=0; element != NULL; i++, element = array[i])
                free(element);

        free(array);
}

static char *strdup(const char *s1)
{
        char *s;
        int len = strlen(s1);

        s = (char *)malloc(len+1);
        if (!s) return NULL;
        strncpy(s, s1, len+1);
        s[len] = '\0';
        return s;
}

@implementation NSArray (DoublePointerConveniences)
- (char **)cStringArray
{
        const char **res;
        int i;
        int len = [self count];

        [self retain];

        res = (const char **)calloc(len + 1, sizeof(char *));

        for (i = 0; i < len; i++)
                {
                /* cString is autoreleased */
                res[i] = strdup([[self objectAtIndex:i] cString]);
                }

        res[len] = (char *)NULL;

        [self release];

        return (char **)res;
}

+ (NSArray *)arrayWithCStringArray:(char **)res
{
        NSMutableArray *theArray = [NSMutableArray array];
        char *element;
        int i = 0;

        element = res[i];

        if (element == NULL)
                {
                return nil;
                }

        while (element != NULL)
                {
                i++;
                [theArray addObject:[NSString stringWithCString:element]];		
//		[theArray addObject:[[[NSString alloc] 
//			initWithCStringNoCopy:element length:strlen(element) freeWhenDone:YES] autorelease]];

                element = res[i];
                }

        return theArray;	
}
@end

@implementation EOFetchSpecification (LDAPUtilities)
- (NSString *)ldapFilterWithValueSeparator:(NSString *)valueSeparator
{
	NSMutableString *filter = [NSMutableString stringWithString:@"("];

	[filter appendFormat:@"&(objectclass=%@)", [self entityName]];

	if ([self qualifier])
		[filter appendString:[[self qualifier] ldapFilterWithValueSeparator:valueSeparator]];

	[filter appendString:@")"];
	return filter;
}
@end

@implementation EOQualifier (LDAPUtilities)
- (NSString *)ldapFilterWithValueSeparator:(NSString *)valueSeparator
{
	return [NSString string];
}
@end

@implementation EONotQualifier (LDAPUtilities)
- (NSString *)ldapFilterWithValueSeparator:(NSString *)valueSeparator
{
	EOQualifier *notQualified = [self qualifier];

	return [NSString stringWithFormat:@"(!(%@))", [notQualified ldapFilterWithValueSeparator:valueSeparator]];
}
@end

@implementation EOAndQualifier (LDAPUtilities)
- (NSString *)ldapFilterWithValueSeparator:(NSString *)valueSeparator
{
	int i;
	NSArray *qualifiers = [self qualifiers];
	NSMutableString *filter = [NSMutableString stringWithString:@"(&"];
	int count = [qualifiers count];

	for (i=0; i<count; i++)
		[filter appendString:[[qualifiers objectAtIndex:i] ldapFilterWithValueSeparator:valueSeparator]];

	[filter appendString:@")"];

	return filter;
}
@end

@implementation EOOrQualifier (LDAPUtilities)
- (NSString *)ldapFilterWithValueSeparator:(NSString *)valueSeparator
{
	int i;
	NSArray *qualifiers = [self qualifiers];
	NSMutableString *filter = [NSMutableString stringWithString:@"(|"];
	int count = [qualifiers count];

	for (i=0; i<count; i++)
		[filter appendString:[[qualifiers objectAtIndex:i] ldapFilterWithValueSeparator:valueSeparator]];

	[filter appendString:@")"];

	return filter;
}
@end

/* Big problem here.  If the EOModel has attributes with names different from
   the column name (eg. "FullName" for "cn"), the qualifier is given that key,
   which would really screw up the search.  I need some way to get the EOModel,
   so I can get the attributes and determine the correct key. */
@implementation EOKeyValueQualifier (LDAPUtilities)
- (NSString *)ldapFilterWithValueSeparator:(NSString *)valueSeparator
{
	NSString *prefix, *key, *selector, *value, *suffix;
	NSArray *values;
	NSMutableString *filter;
	int count, i;

	key = [NSString stringWithString:[self key]];
	if ([key isEqualToString:@"dn"] || [key isEqualToString:@"parent"] || [key isEqualToString:@"rdn"])
        	return nil;
	prefix = [NSString stringWithString:@"("];
	selector = [NSString stringWithString:[EOQualifier stringForOperatorSelector:[self selector]]];
	suffix = [NSString stringWithString:@")"];

	if ([self selector] == EOQualifierOperatorNotEqual) {
		prefix = [NSString stringWithString:@"(!("];
		selector = [NSString stringWithString:@"="];
		suffix = [NSString stringWithString:@"))"];
	} else if ([self selector] == EOQualifierOperatorLessThan) {
		selector = [NSString stringWithString:@"<="];
	} else if ([self selector] == EOQualifierOperatorGreaterThan) {
			selector = [NSString stringWithString:@">="];
	} else if (([self selector] == EOQualifierOperatorLike) ||
			([self selector] == EOQualifierOperatorCaseInsensitiveLike) ||
			([self selector] == EOQualifierOperatorContains)) {
		selector = [NSString stringWithString:@"="];
	}

//	if ( ([self selector] == EOQualifierOperatorLike) || ([self selector] == EOQualifierOperatorCaseInsensitiveLike) || ([self selector] == EOQualifierOperatorContains))
//        	selector = [NSString stringWithString:@"="];
//	else
//        	selector = [NSString stringWithString:[EOQualifier stringForOperatorSelector:[self selector]]];
	value = [NSString stringWithString:[self value]];
	if ([value isEqualToString:@""])
		return nil;

	values = [value componentsSeparatedByString:valueSeparator];
	count = [values count];
	if (count == 1)
        	return [NSString stringWithFormat:@"%@%@%@%@%@", prefix, key, selector, value, suffix];

	filter = [NSMutableString stringWithString:@"(&"];
	for (i = 0; i < count; i++)
		[filter appendFormat:@"%@%@%@%@%@", prefix, key, selector, [values objectAtIndex:i], suffix];
	[filter appendString:@")"];
	return filter;
}
@end